
/** \class Efficiency
 *
 *  Selects candidates from the InputArray according to the efficiency formula.
 *
 *  $Date: 2012-11-18 15:57:08 +0100 (Sun, 18 Nov 2012) $
 *  $Revision: 814 $
 *
 *
 *  \author P. Demin - UCL, Louvain-la-Neuve
 *
 */

#include "modules/Efficiency.h"

#include "classes/DelphesClasses.h"
#include "classes/DelphesFactory.h"
#include "classes/DelphesFormula.h"

#include "ExRootAnalysis/ExRootResult.h"
#include "ExRootAnalysis/ExRootFilter.h"
#include "ExRootAnalysis/ExRootClassifier.h"

#include "TMath.h"
#include "TString.h"
#include "TFormula.h"
#include "TRandom3.h"
#include "TObjArray.h"
#include "TDatabasePDG.h"
#include "TLorentzVector.h"

#include <algorithm> 
#include <stdexcept>
#include <iostream>
#include <sstream>

using namespace std;

//------------------------------------------------------------------------------

Efficiency::Efficiency() :
  fItInputArray(0)
{
}

//------------------------------------------------------------------------------

Efficiency::~Efficiency()
{
}

//------------------------------------------------------------------------------

void Efficiency::Init()
{
  map< Int_t, DelphesFormula * >::iterator itFormulaMap;
  ExRootConfParam param;
  DelphesFormula *formula;
  Int_t i, size;

  // read efficiency formulas
  param = GetParam("EfficiencyFormula");
  size = param.GetSize();
  
  fFormulaMap.clear();
  for(i = 0; i < size/2; ++i)
  {
    formula = new DelphesFormula;
    formula->Compile(param[i*2 + 1].GetString());

    fFormulaMap[param[i*2].GetInt()] = formula;
  }

  // set default efficiency formula
  itFormulaMap = fFormulaMap.find(0);
  if(itFormulaMap == fFormulaMap.end())
  {
    formula = new DelphesFormula;
    formula->Compile("1.0");

    fFormulaMap[0] = formula;
  }

  // import array with output from filter/classifier module

  fInputArray = ImportArray(GetString("InputArray", "ParticlePropagator/candidates"));
  fItInputArray = fInputArray->MakeIterator();

  // create output arrays

  fOutputArray = ExportArray(GetString("OutputArray", "candidates"));

}

//------------------------------------------------------------------------------

void Efficiency::Finish()
{
  map< Int_t, DelphesFormula * >::iterator itFormulaMap;
  DelphesFormula *formula;
  
  if(fItInputArray) delete fItInputArray;
  
  for(itFormulaMap = fFormulaMap.begin(); itFormulaMap != fFormulaMap.end(); ++itFormulaMap)
  {
    formula = itFormulaMap->second;
    if(formula) delete formula;
  }
}

//------------------------------------------------------------------------------

void Efficiency::Process()
{ 
  Candidate *candidate;
  Double_t pt, eta, phi;
  Int_t pdgCode; 
  map< Int_t, DelphesFormula * >::iterator itFormulaMap;
  DelphesFormula *formula;

  fItInputArray->Reset();
  while((candidate = static_cast<Candidate*>(fItInputArray->Next())))
  {
    const TLorentzVector &candidatePosition = candidate->Position;
    const TLorentzVector &candidateMomentum = candidate->Momentum;
    pdgCode = TMath::Abs(candidate->PID);
    eta = candidatePosition.Eta();
    phi = candidatePosition.Phi();
    pt = candidateMomentum.Pt();

    // find an efficency formula
    itFormulaMap = fFormulaMap.find(pdgCode);
    if(itFormulaMap == fFormulaMap.end())
    {
      itFormulaMap = fFormulaMap.find(0);
    }
    formula = itFormulaMap->second;

    // apply an efficency formula
    if(gRandom->Uniform() > formula->Eval(pt, eta)) continue;
    
    fOutputArray->Add(candidate);
  }
}

//------------------------------------------------------------------------------
