
#include "modules/MagneticField.h"

#include "classes/DelphesClasses.h"
#include "classes/DelphesFactory.h"
#include "classes/DelphesFormula.h"

#include "ExRootAnalysis/ExRootResult.h"
#include "ExRootAnalysis/ExRootFilter.h"
#include "ExRootAnalysis/ExRootClassifier.h"

#include "TMath.h"
#include "TString.h"
#include "TFormula.h"
#include "TRandom3.h"
#include "TObjArray.h"
#include "TDatabasePDG.h"
#include "TLorentzVector.h"

#include <algorithm> 
#include <stdexcept>
#include <iostream>
#include <sstream>

using namespace std;

//------------------------------------------------------------------------------

MagneticField::MagneticField() :
  fItInputArray(0)
{
}

//------------------------------------------------------------------------------

MagneticField::~MagneticField()
{
}

//------------------------------------------------------------------------------

void MagneticField::Init()
{
  fMaxIterations = GetInt("MaxIterations", 10000);
  fRadius = GetDouble("Radius", 1.0);
  fRadius2 = fRadius*fRadius;
  fLength = GetDouble("Length", 5.0);
  fBx = GetDouble("Bx", 0.0);
  fBy = GetDouble("By", 0.0);
  fBz = GetDouble("Bz", 0.0);
  
  fOverflowCounter = 0;

  // import array with output from filter/classifier module

  fInputArray = ImportArray(GetString("InputArray", "candidates"));
  fItInputArray = fInputArray->MakeIterator();

  // create output arrays

  fOutputArray = ExportArray("candidates");

}

//------------------------------------------------------------------------------

void MagneticField::Finish()
{
  if(fOverflowCounter > 0)
  {
    cout << "MagneticField OverflowCounter: " << fOverflowCounter << " cases" << endl;
  }
}

//------------------------------------------------------------------------------

void MagneticField::Process()
{
  Candidate *candidate;
  Double_t px, py, pz, pt, p;
  Double_t vx, vy, vz, m, q, qm;
  Double_t x, y, z, ax, ay, az, dt, r2;
  Double_t xold, yold, zold;
  Double_t vtratio, vtold;
  Double_t theta, etaCalo, phiCalo;
  Int_t k;
    
  fItInputArray->Reset();
  while((candidate = static_cast<Candidate*>(fItInputArray->Next())))
  {
    x = candidate->Position.X();
    y = candidate->Position.Y();
    z = candidate->Position.Z();
  
    px = candidate->Momentum.Px() / 0.003;
    py = candidate->Momentum.Py() / 0.003;
    pz = candidate->Momentum.Pz() / 0.003;
    pt = candidate->Momentum.Pt() / 0.003;
    p = TMath::Hypot(pz, pt);
 
    m = candidate->Mass;
    q = candidate->Charge;

    vx = px/m;
    vy = py/m;
    vz = pz/m;
    qm = q/m;
 
//double v = sqrt(vx*vx + vy*vy + vz*vz)/3E8;
//cout << "v = " << v;
//double gamma = 1./sqrt(1-v*v);
//cout << "gamma = " << gamma << endl;

    ax = qm*(fBz*vy - fBy*vz);
    ay = qm*(fBx*vz - fBz*vx);
    az = qm*(fBy*vx - fBx*vy);
    dt = 1.0/p;
    
    if(pt < 266 && vz < 0.0012) dt = TMath::Abs(0.001/vz); // ?????
  
    xold = x;
    yold = y;
    zold = z;
 
    vtold = pt/m; // = sqrt(vx*vx + vy*vy);
 
    k = 0;
    vtratio = 0.0;

    r2 = 0; // will be x*x + y*y

    while(k < fMaxIterations)
    {
      ++k;
    
      vx += ax*dt;
      vy += ay*dt;
      vz += az*dt;
    
      vtratio = vtold/sqrt(vx*vx+vy*vy);
      vx *= vtratio;
      vy *= vtratio;
    
      ax = qm*(fBz*vy - fBy*vz);
      ay = qm*(fBx*vz - fBz*vx);
      az = qm*(fBy*vx - fBx*vy);
    
      x += vx*dt;
      y += vy*dt;
      z += vz*dt;
      r2 = x*x + y*y;
    
      if(r2 > fRadius2)
      { 
        x /= r2/fRadius2; 
        y /= r2/fRadius2; 
        break;
      }
      if(TMath::Abs(z) > fLength) break;
    
      xold = x;
      yold = y;
      zold = z;
    } // while loop
  
    if(k == fMaxIterations) ++fOverflowCounter;
    theta = 0.0; 
    if(x != 0 && y != 0 && z != 0)
    {
      theta = TMath::ATan2(TMath::Sqrt(r2), z);
      etaCalo = -TMath::Log(TMath::Tan(theta/2.0));
      phiCalo = TMath::ATan2(y, x);
      
      candidate = static_cast<Candidate*>(candidate->Clone());
      
      candidate->Position.SetXYZT(x, y ,z, 0.0);

      candidate->Momentum.SetPtEtaPhiE(candidate->Momentum.Pt(), etaCalo, phiCalo, candidate->Momentum.E());
      
      fOutputArray->Add(candidate);
    }
  }
}

//------------------------------------------------------------------------------
