
/** \class TauJetFinder
 *
 *  Finds hadronic tau decays.
 *
 *  $Date: 2012-11-18 15:57:08 +0100 (Sun, 18 Nov 2012) $
 *  $Revision: 814 $
 *
 *
 *  \author P. Demin - UCL, Louvain-la-Neuve
 *
 */

#include "modules/TauJetFinder.h"

#include "classes/DelphesClasses.h"
#include "classes/DelphesFactory.h"
#include "classes/DelphesFormula.h"

#include "ExRootAnalysis/ExRootResult.h"
#include "ExRootAnalysis/ExRootFilter.h"
#include "ExRootAnalysis/ExRootClassifier.h"

#include "TMath.h"
#include "TString.h"
#include "TFormula.h"
#include "TRandom3.h"
#include "TObjArray.h"
#include "TDatabasePDG.h"
#include "TLorentzVector.h"

#include <algorithm> 
#include <stdexcept>
#include <iostream>
#include <sstream>

using namespace std;

//------------------------------------------------------------------------------

TauJetFinder::TauJetFinder() :
  fItTrackInputArray(0), fItTowerInputArray(0), fItJetInputArray(0)
{
}

//------------------------------------------------------------------------------

TauJetFinder::~TauJetFinder()
{
}

//------------------------------------------------------------------------------

void TauJetFinder::Init()
{
  fTrackDeltaR = GetDouble("TrackDeltaR", 0.4);
  fTowerDeltaR = GetDouble("TowerDeltaR", 0.15);

  fTrackPTMin = GetDouble("TrackPTMin", 2.0);
  fTowerETMin = GetDouble("TowerETMin", 1.0);
  fJetPTMin = GetDouble("JetPTMin", 10.0);

  fEnergyFractionMin = GetDouble("EnergyFractionMin", 0.95);  

  // import input array(s)

  fTrackInputArray = ImportArray(GetString("TrackInputArray", "TrackFinder/tracks"));
  fItTrackInputArray = fTrackInputArray->MakeIterator();

  fTowerInputArray = ImportArray(GetString("TowerInputArray", "Calorimeter/towers"));
  fItTowerInputArray = fTowerInputArray->MakeIterator();
  
  fJetInputArray = ImportArray(GetString("JetInputArray", "FastJetFinder/jets"));
  fItJetInputArray = fJetInputArray->MakeIterator();

  // create output array

  fOutputArray = ExportArray(GetString("OutputArray", "taus"));
}

//------------------------------------------------------------------------------

void TauJetFinder::Finish()
{
  if(fItJetInputArray) delete fItJetInputArray;
  if(fItTowerInputArray) delete fItTowerInputArray;
  if(fItTrackInputArray) delete fItTrackInputArray;
}

//------------------------------------------------------------------------------

void TauJetFinder::Process()
{
  Candidate *jet, *track, *tower;
  TLorentzVector jetMomentum;
  Int_t trackCounter, charge;
  Double_t towerEnergy;
  
  // loop over all input jets
  fItJetInputArray->Reset();
  while((jet = static_cast<Candidate*>(fItJetInputArray->Next())))
  {
    const TLorentzVector &jetMomentum = jet->Momentum;
    
    if(jetMomentum.Pt() <= fJetPTMin) continue;

    // loop over all tracks
    trackCounter = 0;
    charge = 0;
    fItTrackInputArray->Reset();
    while((track = static_cast<Candidate*>(fItTrackInputArray->Next())))
    {
      if(track->Momentum.Pt() <= fTrackPTMin) continue;
      if(jetMomentum.DeltaR(track->Momentum) <= fTrackDeltaR)
      {
        ++trackCounter;
        charge += track->Charge;
      }
    }
    
    if(trackCounter != 1 && trackCounter != 3) continue;

    // loop over all towers
    towerEnergy = 0.0;
    fItTowerInputArray->Reset();
    while((tower = static_cast<Candidate*>(fItTowerInputArray->Next())))
    {
      if(tower->Momentum.Pt() <= fTowerETMin) continue;
      if(jetMomentum.DeltaR(tower->Momentum) <= fTowerDeltaR)
      {
        towerEnergy += tower->Momentum.E();
      }
    }
    
    if(towerEnergy/jetMomentum.E() <= fEnergyFractionMin) continue;
    
    jet->Charge = charge;

    fOutputArray->Add(jet);
  }
}

//------------------------------------------------------------------------------
