#include <stdexcept>
#include <iostream>
#include <sstream>

#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <rpc/xdr.h>

#include "TROOT.h"
#include "TApplication.h"

#include "TFile.h"
#include "TObjArray.h"
#include "TStopwatch.h"
#include "TDatabasePDG.h"
#include "TParticlePDG.h"
#include "TLorentzVector.h"

#include "modules/Delphes.h"
#include "classes/DelphesClasses.h"
#include "classes/DelphesFactory.h"

#include "ExRootAnalysis/ExRootTreeWriter.h"
#include "ExRootAnalysis/ExRootTreeBranch.h"
#include "ExRootAnalysis/ExRootProgressBar.h"

using namespace std;

//------------------------------------------------------------------------------

class STDHEPConverter
{
public:
  enum STDHEPBlock {GENERIC = 0, 
                    FILEHEADER = 1, 
                    EVENTTABLE = 2, 
                    EVENTHEADER = 4,
                    MCFIO_STDHEP = 101,
                    MCFIO_STDHEPBEG = 106,
                    MCFIO_STDHEPEND = 107,
                    MCFIO_STDHEP4 = 201};

  STDHEPConverter();
  ~STDHEPConverter();
  
  void SetXDR(XDR *xdrs) {fXDR = xdrs;}

  void ReadFileHeader();
  void ReadBlock();

  UInt_t GetEntries() {return fEntries;}; 
  Int_t GetBlockType() {return fBlockType;};

  void AnalyzeEvent(ExRootTreeBranch *branch,
                    TStopwatch *readStopWatch, TStopwatch *procStopWatch);

  void AnalyzeParticles(DelphesFactory *factory, TObjArray *particleOutputArray,
                        TObjArray *candidateOutputArray, TObjArray *partonOutputArray);
                        
  void FinalizeParticles(TObjArray *particleOutputArray);

private:

  void SkipBytes(u_int size);
  void SkipArray(u_int elsize);

  void ReadEventTable();
  void ReadSTDHEPHeader();
  void ReadSTDCM1();
  void ReadSTDHEP();
  void ReadSTDHEP4();

  void FixDaughters(TObjArray *particleOutputArray, Int_t indexParticle, Int_t indexMother);

  XDR *fXDR;
  
  char *fBuffer;
  
  UInt_t fEntries;
  Int_t fBlockType, fEventNumber, fEventSize;
  Double_t fWeight, fAlphaQCD, fAlphaQED;
  
  u_int fScaleSize;
  double fScale[10];
};

//---------------------------------------------------------------------------

STDHEPConverter::STDHEPConverter() :
  fXDR(0), fBuffer(0), fEntries(0)
{
  fBuffer = new char[10000*96 + 24];
}

//---------------------------------------------------------------------------

STDHEPConverter::~STDHEPConverter()
{
  if(fBuffer) delete fBuffer;
}

//---------------------------------------------------------------------------

void STDHEPConverter::SkipBytes(u_int size)
{
  xdr_opaque(fXDR, fBuffer, size);
}

//---------------------------------------------------------------------------

void STDHEPConverter::SkipArray(u_int elsize)
{
  u_int size;
  xdr_u_int(fXDR, &size);
  SkipBytes(size*elsize);
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadFileHeader()
{
  UInt_t i;
  enum STDHEPVersion {UNKNOWN, V1, V2, V21} version;

  xdr_int(fXDR, &fBlockType);
  if (fBlockType != FILEHEADER)
  {
    throw runtime_error("Header block not found. File is probably corrupted.");
  }

  SkipBytes(4);

  // version
  xdr_string(fXDR, &fBuffer, 100);
  if(strnlen(fBuffer, 2) < 2) version = UNKNOWN;
  else if(fBuffer[0] == '1') version = V1;
  else if(strncmp(fBuffer, "2.01", 4) == 0) version = V21;
  else if(fBuffer[0] == '2') version = V2;
  else version = UNKNOWN;
  
  if (version == UNKNOWN) 
  {
    throw runtime_error("Unknown file format version.");
  }

  SkipArray(1);
  SkipArray(1);
  SkipArray(1);

  if (version == V21)
  {
    SkipArray(1);
  }

  // Expected number of events
  SkipBytes(4);

  // Number of events
  xdr_u_int(fXDR, &fEntries);

  SkipBytes(8);

  // Number of blocks
  UInt_t nBlocks = 0;
  xdr_u_int(fXDR, &nBlocks);

  // Number of NTuples
  UInt_t nNTuples = 0;
  if(version != V1)
  {
    xdr_u_int(fXDR, &nNTuples);
  }

  if(nNTuples != 0)
  {
    throw runtime_error("Files containing n-tuples are not supported.");
  }

  // Processing blocks extraction 
  if(nBlocks != 0)
  {
    SkipArray(4);

    for(i = 0; i < nBlocks; i++)
    {
      SkipArray(1);
    }
  }
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadBlock()
{
  xdr_int(fXDR, &fBlockType);

  SkipBytes(4);

  xdr_string(fXDR, &fBuffer, 100);

  if(fBlockType == EVENTTABLE) ReadEventTable();
  else if(fBlockType == EVENTHEADER) ReadSTDHEPHeader();
  else if(fBlockType == MCFIO_STDHEPBEG ||
          fBlockType == MCFIO_STDHEPEND) ReadSTDCM1();
  else if(fBlockType == MCFIO_STDHEP) ReadSTDHEP();
  else if(fBlockType == MCFIO_STDHEP4) ReadSTDHEP4();
  else
  {
    throw runtime_error("Unsupported block type.");
  }
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadEventTable()
{
  SkipBytes(8);

  SkipArray(4);
  SkipArray(4);
  SkipArray(4);
  SkipArray(4);
  SkipArray(4);
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadSTDHEPHeader()
{
  SkipBytes(20);

  UInt_t dimBlocks = 0;
  xdr_u_int(fXDR, &dimBlocks);

  UInt_t dimNTuples = 0;
  if(strncmp(fBuffer, "2.00", 4))
  {
    SkipBytes(4);
    xdr_u_int(fXDR, &dimNTuples);
  }

  // Processing blocks extraction 
  if(dimBlocks > 0)
  {
    SkipArray(4);
    SkipArray(4);
  }

  // Processing blocks extraction
  if(dimNTuples > 0 && strncmp(fBuffer, "2.00", 4))
  {
    SkipArray(4);
    SkipArray(4);
  }
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadSTDCM1()
{
  // skip 5*4 + 2*8 = 36 bytes
  SkipBytes(36);

  if((strncmp(fBuffer, "1.", 2) == 0) || (strncmp(fBuffer, "2.", 2) == 0) ||
     (strncmp(fBuffer, "3.", 2) == 0) || (strncmp(fBuffer, "4.", 2) == 0) ||
     (strncmp(fBuffer, "5.00", 4) == 0))
  {
    return;
  }

  SkipArray(1);
  SkipArray(1);

  if(strncmp(fBuffer, "5.01", 4) == 0)
  {
    return;
  }

  SkipBytes(4);
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadSTDHEP()
{
  UInt_t idhepSize, isthepSize, jmohepSize, jdahepSize, phepSize, vhepSize;

  // Extracting the event number
  xdr_int(fXDR, &fEventNumber);  

  // Extracting the number of particles
  xdr_int(fXDR, &fEventSize);  

  // 4*n + 4*n + 8*n + 8*n + 40*n + 32*n +
  // 4 + 4 + 4 + 4 + 4 + 4 = 96*n + 24

  xdr_opaque(fXDR, fBuffer, 96*fEventSize + 24);

  idhepSize = ntohl(*(UInt_t*)(fBuffer));
  isthepSize = ntohl(*(UInt_t*)(fBuffer + 4*1 + 4*1*fEventSize));
  jmohepSize = ntohl(*(UInt_t*)(fBuffer + 4*2 + 4*2*fEventSize));
  jdahepSize = ntohl(*(UInt_t*)(fBuffer + 4*3 + 4*4*fEventSize));
  phepSize = ntohl(*(UInt_t*)(fBuffer + 4*4 + 4*6*fEventSize));
  vhepSize = ntohl(*(UInt_t*)(fBuffer + 4*5 + 4*16*fEventSize));

  if(fEventSize < 0 ||
     fEventSize != (Int_t)idhepSize      || fEventSize != (Int_t)isthepSize     || 
     (2*fEventSize) != (Int_t)jmohepSize || (2*fEventSize) != (Int_t)jdahepSize ||
     (5*fEventSize) != (Int_t)phepSize   || (4*fEventSize) != (Int_t)vhepSize)
  {
    throw runtime_error("Inconsistent size of arrays. File is probably corrupted.");
  }
  
  fWeight = 1.0;
  fAlphaQED = 0.0;
  fAlphaQCD = 0.0;
  fScaleSize = 0;
  memset(fScale, 0, 10*sizeof(double));
}

//---------------------------------------------------------------------------

void STDHEPConverter::ReadSTDHEP4()
{
  u_int number;

  ReadSTDHEP();

  // Extracting the event weight
  xdr_double(fXDR, &fWeight);  

  // Extracting alpha QED
  xdr_double(fXDR, &fAlphaQED);

  // Extracting alpha QCD
  xdr_double(fXDR, &fAlphaQCD);

  // Extracting the event scale
  xdr_u_int(fXDR, &fScaleSize);
  for(number = 0; number < fScaleSize; ++number)
  {
    xdr_double(fXDR, &fScale[number]);
  }

  SkipArray(8);
  SkipArray(4);

  SkipBytes(4);
}

//---------------------------------------------------------------------------

void STDHEPConverter::AnalyzeEvent(ExRootTreeBranch *branch,
                                   TStopwatch *readStopWatch, TStopwatch *procStopWatch)
{
  LHEFEvent *element;

  element = static_cast<LHEFEvent *>(branch->NewEntry());

  element->Number = fEventNumber;
  
  element->ProcessID = 0;

  element->Weight = fWeight;
  element->ScalePDF = fScale[0];
  element->AlphaQED = fAlphaQED;
  element->AlphaQCD = fAlphaQCD;
  
  element->ReadTime = readStopWatch->RealTime();
  element->ProcTime = procStopWatch->RealTime();
}

//---------------------------------------------------------------------------

void STDHEPConverter::AnalyzeParticles(DelphesFactory *factory,
                                       TObjArray *particleOutputArray,
                                       TObjArray *candidateOutputArray,
                                       TObjArray *partonOutputArray)
{
  Candidate *candidate;

  TLorentzVector momentum;

  TDatabasePDG *pdg = TDatabasePDG::Instance();
  TParticlePDG *pdgParticle;

  Int_t number;
  Int_t pid, status, m1, m2, d1, d2;
  Double_t px, py, pz, e, m;  
  Double_t x, y, z, t;  
  
  XDR xdrs[6];
  xdrmem_create(&xdrs[0], fBuffer + 4*1, 4*fEventSize, XDR_DECODE);
  xdrmem_create(&xdrs[1], fBuffer + 4*2 + 4*1*fEventSize, 4*fEventSize, XDR_DECODE);
  xdrmem_create(&xdrs[2], fBuffer + 4*3 + 4*2*fEventSize, 8*fEventSize, XDR_DECODE);
  xdrmem_create(&xdrs[3], fBuffer + 4*4 + 4*4*fEventSize, 8*fEventSize, XDR_DECODE);
  xdrmem_create(&xdrs[4], fBuffer + 4*5 + 4*6*fEventSize, 40*fEventSize, XDR_DECODE);
  xdrmem_create(&xdrs[5], fBuffer + 4*6 + 4*16*fEventSize, 32*fEventSize, XDR_DECODE);

  for(number = 0; number < fEventSize; ++number)
  {
    xdr_int(&xdrs[0], &status);
    xdr_int(&xdrs[1], &pid);
    xdr_int(&xdrs[2], &m1);
    xdr_int(&xdrs[2], &m2);
    xdr_int(&xdrs[3], &d1);
    xdr_int(&xdrs[3], &d2);

    xdr_double(&xdrs[4], &px);
    xdr_double(&xdrs[4], &py);
    xdr_double(&xdrs[4], &pz);
    xdr_double(&xdrs[4], &e);
    xdr_double(&xdrs[4], &m);
    
    xdr_double(&xdrs[5], &x);
    xdr_double(&xdrs[5], &y);
    xdr_double(&xdrs[5], &z);
    xdr_double(&xdrs[5], &t);

    candidate = factory->NewCandidate();
    
    candidate->PID = pid;

    candidate->Status = status;

    candidate->M1 = m1 - 1;
    candidate->M2 = m2 - 1;

    candidate->D1 = d1 - 1;
    candidate->D2 = d2 - 1;

    pdgParticle = pdg->GetParticle(pid);
    candidate->Charge = pdgParticle ? Int_t(pdgParticle->Charge()/3.0) : -999;
    candidate->Mass = pdgParticle ? pdgParticle->Mass() : -999.9;

    candidate->Momentum.SetPxPyPzE(px, py, pz, e);
    
    candidate->Position.SetXYZT(x, y, z, t);

    particleOutputArray->Add(candidate);

    if(!pdgParticle) continue;

    if(status == 1)
    {
      candidateOutputArray->Add(candidate);
    }
    else if(status == 2)
    {
      partonOutputArray->Add(candidate);
    }
  }
}

//---------------------------------------------------------------------------

void STDHEPConverter::FinalizeParticles(TObjArray *particleOutputArray)
{
  Int_t indexParticle, indexMother;
  Candidate *candidate;
  Int_t status, m1, m2, d1;

  for(indexParticle = 0; indexParticle < particleOutputArray->GetEntriesFast(); ++indexParticle)
  {
    candidate = static_cast<Candidate *>(particleOutputArray->At(indexParticle));
    status = candidate->Status;
    m1 = candidate->M1;
    m2 = candidate->M2;
    d1 = candidate->D1;

    if(m1 < 0) continue;
       
    if(m2 < 0)
    {
      FixDaughters(particleOutputArray, indexParticle, m1);
    }
    else
    {
      for(indexMother = m1; indexMother <= m2; ++indexMother)
      {
        FixDaughters(particleOutputArray, indexParticle, indexMother);
      }
    }
  }
}

//---------------------------------------------------------------------------

void STDHEPConverter::FixDaughters(TObjArray *particleOutputArray, Int_t indexParticle, Int_t indexMother)
{
  Candidate *mother = 0;
  Int_t status, m1, m2, d1, d2;

  mother = static_cast<Candidate *>(particleOutputArray->At(indexMother));

  if(!mother) return;

  status = mother->Status;
  m1 = mother->M1;
  m2 = mother->M2;
  d1 = mother->D1;
  d2 = mother->D2;

  // skip beam particles
  if(m1 < 0 && m2 < 0) return;

  // skip final state particles
  if(status != 3 && status != 2) return;
  
  if(d1 < 0 && d2 < 0)
  {
    mother->D1 = indexParticle;
    mother->D2 = indexParticle;
  }
  else if(d1 < indexParticle && d2 + 1 == indexParticle)
  {
    mother->D2 = indexParticle;
  }
  else if(d1 < indexParticle)
  {
    mother->D1 = indexParticle;
    mother->D2 = indexParticle;
  }
} 

//---------------------------------------------------------------------------

static bool interrupted = false;

void SignalHandler(int sig)
{
	interrupted = true;
}

//---------------------------------------------------------------------------

int main(int argc, char *argv[])
{
  char appName[] = "DelphesSTDHEP";
  stringstream message;
  FILE *inputFile = 0;
  XDR xdrs;
  TFile *outputFile = 0;
  TStopwatch readStopWatch, procStopWatch;
  ExRootTreeWriter *treeWriter = 0;
  ExRootTreeBranch *branchEvent = 0;
  ExRootConfReader *confReader = 0;
  Delphes *modularDelphes = 0;
  DelphesFactory *factory = 0;
  TObjArray *particleOutputArray = 0, *candidateOutputArray = 0, *partonOutputArray = 0;
  STDHEPConverter *converter = 0;
  Int_t i, entry, allEntries;
  Long64_t length, eventCounter;

  if(argc < 3)
  {
    cout << " Usage: " << appName << " config_file" << " output_file" << " [input_file(s)]" << endl;
    cout << " config_file - configuration file in Tcl format," << endl;
    cout << " output_file - output file in ROOT format," << endl;
    cout << " input_file(s) - input file(s) in STDHEP format," << endl;
    cout << " with no input_file, or when input_file is -, read standard input." << endl;
    return 1;
  }

  signal(SIGINT, SignalHandler);

  gROOT->SetBatch();

  int appargc = 1;
  char *appargv[] = {appName};
  TApplication app(appName, &appargc, appargv);

  try
  {
    outputFile = TFile::Open(argv[2], "RECREATE");

    if(outputFile == NULL)
    {
      message << "can't open " << argv[2];
      throw runtime_error(message.str());
    }

    treeWriter = new ExRootTreeWriter(outputFile, "Delphes");

    branchEvent = treeWriter->NewBranch("Event", LHEFEvent::Class());

    confReader = new ExRootConfReader;
    confReader->ReadFile(argv[1]);
    
    modularDelphes = new Delphes("Delphes");
    modularDelphes->SetConfReader(confReader);
    modularDelphes->SetTreeWriter(treeWriter);
    
    factory = modularDelphes->GetFactory();
    particleOutputArray = modularDelphes->ExportArray("particles");
    candidateOutputArray = modularDelphes->ExportArray("candidates");
    partonOutputArray = modularDelphes->ExportArray("partons");

    converter = new STDHEPConverter();

    modularDelphes->InitTask();

    i = 3;
    do
    {
      if(interrupted) break;

      if(i == argc || strcmp(argv[i], "-") == 0)
      {
        cout << "** Reading standard input" << endl;
        inputFile = stdin;
        length = -1;
      }
      else
      {
        cout << "** Reading " << argv[i] << endl;
        inputFile = fopen(argv[i], "r");

        if(inputFile == NULL)
        {
          message << "can't open " << argv[i];
          throw runtime_error(message.str());
        }

        fseek(inputFile, 0L, SEEK_END);
        length = ftell(inputFile);
        fseek(inputFile, 0L, SEEK_SET);
      }

      xdrstdio_create(&xdrs, inputFile, XDR_DECODE);
      converter->SetXDR(&xdrs);

      converter->ReadFileHeader();
    
      allEntries = converter->GetEntries();
      
      if(allEntries <= 0) continue;

      ExRootProgressBar progressBar(length);
  
      // Loop over all objects
      eventCounter = 0;
      entry = 0;
      treeWriter->Clear();
      modularDelphes->Clear();
      readStopWatch.Start();
      while(entry < allEntries && !feof(inputFile) && !interrupted)
      {
        converter->ReadBlock();
        if(converter->GetBlockType() == STDHEPConverter::EVENTTABLE ||
           converter->GetBlockType() == STDHEPConverter::EVENTHEADER) continue;

        if(converter->GetBlockType() == STDHEPConverter::MCFIO_STDHEP)
        {
          ++eventCounter;

          converter->AnalyzeParticles(factory, particleOutputArray, candidateOutputArray, partonOutputArray); 
          converter->FinalizeParticles(particleOutputArray); 

          readStopWatch.Stop();
          procStopWatch.Start();
          modularDelphes->ProcessTask();
          procStopWatch.Stop();

          converter->AnalyzeEvent(branchEvent, &readStopWatch, &procStopWatch);
    
          treeWriter->Fill();

          treeWriter->Clear();
          modularDelphes->Clear();

          readStopWatch.Start();
        }
        progressBar.Update(ftell(inputFile), eventCounter);
        ++entry;
      }

      fseek(inputFile, 0L, SEEK_END);
      progressBar.Update(ftell(inputFile), eventCounter, kTRUE);
      progressBar.Finish();

      if(inputFile != stdin) fclose(inputFile);

      ++i;
    }
    while(i < argc);

    modularDelphes->FinishTask();
    treeWriter->Write();

    cout << "** Exiting..." << endl;

    delete converter;
    delete modularDelphes;
    delete confReader;
    delete treeWriter;
    delete outputFile;
    
    return 0;
  }
  catch(runtime_error &e)
  {
    if(inputFile) fclose(inputFile);
    if(treeWriter) delete treeWriter;
    if(outputFile) delete outputFile;
    cerr << "** ERROR: " << e.what() << endl;
    return 1;
  }
}

